/// \file PinballGame.ino
/// \brief Arduino program demonstrating essential real-time logic for a custom pinball machine.

/// \copyright No copyright, 2016-2017, Garth Zeglin.  This file is explicitly placed in the public domain.

/// \details This example is intended as a starting point for developing a
///          custom pinball controller, demonstrating the use of state machines
///          to implement both I/O and game logic.
///
/// This example is written using Arduino C++ conventions:
///
///  1. state variables are global
///  2. global C++ objects are statically declared and initialized
///  3. the sketch is divided into multiple .ino files which form one compilation unit
///  4. C++ classes in .cpp files are separately compiled

/// This example assumes that several open-source libraries have been installed in the Arduino IDE:
///  1. Adafruit-GFX	general graphics support
///  2. Adafruit-HT1632	LED matrix display driver
///  3. Adafruit-WS2801	LED strand driver

/****************************************************************/
// Library imports.
#include "SPI.h"
#include "Adafruit_GFX.h"
#include "Adafruit_HT1632.h"
#include "Adafruit_WS2801.h"

// Forward declarations.
#include "PinballSensor.h"
#include "PopBumper.h"
#include "ToneSpeaker.h"
#include "StrandGraphics.h"
#include "MatrixGraphics.h"
#include "console.h"

/****************************************************************/
/**** Hardware pin assignments **********************************/
/****************************************************************/

// The following pin assignments correspond to the hardware on the PinballShield
// Rev A board.

// Analog inputs.
const int PHOTO1_PIN = A0;  /// photo-interrupter input, value decreases when object present
const int PHOTO2_PIN = A1;  /// photo-interrupter input, value decreases when object present
const int PHOTO3_PIN = A2;  /// photo-interrupter input, value decreases when object present
const int PHOTO4_PIN = A3;  /// photo-interrupter input, value decreases when object present
const int SWITCH1_PIN = A4; /// active-low switch input
const int SWITCH2_PIN = A5; /// active-low switch input

// Digital outputs. D0 and D1 are reserved for use as serial port RX/TX
const int LED1_PIN        = 2;  /// active-low LED output
const int SPEAKER_PIN     = 3;  /// active-high MOSFET-driven speaker output
const int LED2_PIN        = 4;  /// active-low LED output
const int SOLENOID4_PIN   = 5;  /// active-high MOSFET-driven solenoid output
const int SOLENOID3_PIN   = 6;  /// active-high MOSFET-driven solenoid output
const int MATRIX_DATA_PIN = 7;  /// HT1632 LED matrix display data output
const int MATRIX_WR_PIN   = 8;  /// HT1632 LED matrix display clock output
const int SOLENOID2_PIN   = 9;  /// active-high MOSFET-driven solenoid output
const int SOLENOID1_PIN   = 10; /// active-high MOSFET-driven solenoid output
const int STRAND_DATA_PIN = 11; /// WS2801 LED strand data output (yellow wire on strand)
const int MATRIX_CS0_PIN  = 12; /// HT1632 LED matrix display select output
const int STRAND_CLK_PIN  = 13; /// WS2801 LED strand clock output (green wire on strand)

/****************************************************************/
/**** Global variables and constants ****************************/
/****************************************************************/

// The baud rate is the number of bits per second transmitted over the serial port.
const long BAUD_RATE = 115200;

PinballSensor start_switch(SWITCH1_PIN);
PinballSensor bumper_sensor(PHOTO1_PIN);
PinballSensor drain_sensor(PHOTO2_PIN);

PopBumper     bumper(SOLENOID1_PIN);
ToneSpeaker   speaker(SPEAKER_PIN);

// Initialize the LED hardware.
Adafruit_HT1632LEDMatrix matrix = Adafruit_HT1632LEDMatrix(MATRIX_DATA_PIN, MATRIX_WR_PIN, MATRIX_CS0_PIN);
Adafruit_WS2801 strand = Adafruit_WS2801((uint16_t) 25, (uint8_t) STRAND_DATA_PIN, (uint8_t) STRAND_CLK_PIN, WS2801_GRB);

// ==============================================================
// Define the game state variables.

/// Define symbolic values for the main game state machine.
enum state_t { STATE_IDLE, STATE_ATTRACT, STATE_BALL1, STATE_BALL2, STATE_BALL3, STATE_GAME_OVER, NUM_STATES } game_state;

/// A count of the number of microseconds elapsed in the current game state.
long game_state_elapsed;

/// Convenient time constants, in microseconds.
const long FIVE_SECONDS = 5000000;

/// Current score (single player only).
unsigned long score = 0;

/// Attract mode melody.
const unsigned char attract_melody[] = {
  MIDI_C4, EIGHTH,   MIDI_E4, EIGHTH,   MIDI_G4, EIGHTH,   MIDI_C5, EIGHTH,   MIDI_B4, EIGHTH,   MIDI_G4, EIGHTH,
  MIDI_E4, QUARTER,
  MIDI_END
};
/****************************************************************/
/// Update all input state, including periodic sensor sampling, debouncing, and filtering.
void poll_sensor_inputs(unsigned long interval)
{
  start_switch.update(interval);
  bumper_sensor.update(interval);
  drain_sensor.update(interval);
}

/****************************************************************/
/// Update the game state machine, including advancing game state and score and applying modal input-output mappings.
void poll_game_logic(unsigned long interval)
{
  game_state_elapsed += interval;

  switch(game_state) {
    //-----------------------------------------------
  case STATE_IDLE:

    if (start_switch.isTriggered()) {
      game_state_elapsed = 0;
      game_state = STATE_BALL1;
      send_debug_message("entering BALL1");
    }

    if (game_state_elapsed > FIVE_SECONDS) {
      game_state_elapsed = 0;
      game_state = STATE_ATTRACT;
      speaker.start_melody(attract_melody);
      strand_set_fast_animation();
      send_debug_message("entering ATTRACT");
    }
    break;

    //-----------------------------------------------
  case STATE_ATTRACT:
    if (start_switch.isTriggered()) {
      game_state_elapsed = 0;
      game_state = STATE_BALL1;
      score = 0; // reset previous score
      send_debug_message("entering BALL1");
    }

    if (game_state_elapsed > FIVE_SECONDS) {
      game_state_elapsed = 0;
      game_state = STATE_IDLE;
      strand_set_slow_animation();
      send_debug_message("entering IDLE");
    }
    break;

    //-----------------------------------------------
  case STATE_BALL1:
    if (drain_sensor.isTriggered()) {
      game_state_elapsed = 0;
      game_state = STATE_BALL2;
      send_debug_message("entering BALL2");
    }

    if (bumper_sensor.isTriggered()) {
      score++;
      bumper.trigger();
    }

    break;

    //-----------------------------------------------
  case STATE_BALL2:
    if (drain_sensor.isTriggered()) {
      game_state_elapsed = 0;
      game_state = STATE_BALL3;
      strand_set_fast_animation();
      send_debug_message("entering BALL3");
    }

    if (bumper_sensor.isTriggered()) {
      score++;
      bumper.trigger();
    }

    break;

    //-----------------------------------------------
  case STATE_BALL3:
    if (drain_sensor.isTriggered()) {
      game_state_elapsed = 0;
      game_state = STATE_GAME_OVER;
      send_debug_message("entering GAME_OVER");
    }

    if (bumper_sensor.isTriggered()) {
      score++;
      bumper.trigger();
    }

    break;

    //-----------------------------------------------
  case STATE_GAME_OVER:
    if (game_state_elapsed > FIVE_SECONDS) {
      game_state_elapsed = 0;
      game_state = STATE_IDLE;
      strand_set_slow_animation();
      send_debug_message("entering IDLE");
    }
    break;
    //-----------------------------------------------
  default:
    // Any other value of game_state is invalid, so re-enter the IDLE state.
    send_debug_message("Unexpected game_state entered.");
    game_state = STATE_IDLE;
    game_state_elapsed = 0;
    break;
  }

  /****************************************************************/

  // some simple LED animation based on the game state
  switch(game_state) {
  case STATE_IDLE:
    digitalWrite(LED1_PIN, HIGH);
    digitalWrite(LED2_PIN, HIGH);
    break;

  case STATE_ATTRACT:
    digitalWrite(LED1_PIN, (game_state_elapsed % 500000) < 250000);
    digitalWrite(LED2_PIN, (game_state_elapsed % 300000) < 150000);
    break;

  case STATE_BALL1:
  case STATE_BALL2:
  case STATE_BALL3:
    if (game_state_elapsed < 1000000) {
      // if the state was freshly entered, flash a little
      digitalWrite(LED1_PIN, (game_state_elapsed % 100000) < 50000);
      digitalWrite(LED2_PIN, (game_state_elapsed % 100000) < 50000);

    } else {
      // otherwise flash when the bumper is hit
      digitalWrite(LED1_PIN, !bumper.isActive());
      digitalWrite(LED2_PIN, !bumper.isActive());
    }
    break;

  case STATE_GAME_OVER:
    digitalWrite(LED1_PIN, (game_state_elapsed % 200000) < 100000);
    digitalWrite(LED2_PIN, (game_state_elapsed % 200000) < 100000);
    break;

  default:
    // On any invalid value or game_state, do nothing.
    break;
  }
}

/****************************************************************/
/// Update all actuator state, including output pulse timers.
void poll_actuator_outputs(unsigned long interval)
{
  bumper.update(interval);
  speaker.update(interval);
  strand_update(interval);

  // For now, just always show the current score on the LED matrix.  This will
  // evolve into a separate LED matrix animation state machine.
  matrix_show_score(score);
}

/****************************************************************/
// Debugging functions which can be called from user console input.
void user_print_report(void)
{
  send_debug_message("start of debugging report.");
  start_switch.send_debug();
  bumper_sensor.send_debug();
  drain_sensor.send_debug();
  bumper.send_debug();
  speaker.send_debug();
  send_debug_message("end of debugging report.");
}
void user_reset_game(void)
{
  send_debug_message("user game reset.");
  game_state = STATE_IDLE;
  game_state_elapsed = 0;
}

void user_set_game_state(int value)
{
  send_debug_message("user forcing game state.");

  // Cast the integer to state_t, which is really just an integer, but
  // considered a different kind of integer by the compiler.
  if (value >= 0 && value < NUM_STATES) {
    game_state = (state_t) value;
    game_state_elapsed = 0;
  }
}
void user_set_game_score(unsigned long value)
{
  send_debug_message("user forcing game score.");
  score = value;
}
void user_play_melody(void)
{
  send_debug_message("user forcing melody.");
  speaker.start_melody(attract_melody);
}

/****************************************************************/
/**** Standard entry points for Arduino system ******************/
/****************************************************************/

/// Standard Arduino initialization function to configure the system.  This
/// function is called once after reset to initialize the program.
void setup()
{
  // configure the actuator pins as soon as possible.
  pinMode(LED1_PIN     , OUTPUT);
  pinMode(SPEAKER_PIN  , OUTPUT);
  pinMode(LED2_PIN     , OUTPUT);
  pinMode(SOLENOID4_PIN, OUTPUT);
  pinMode(SOLENOID3_PIN, OUTPUT);
  pinMode(SOLENOID2_PIN, OUTPUT);
  pinMode(SOLENOID1_PIN, OUTPUT);

  digitalWrite(LED1_PIN     , HIGH); // off
  digitalWrite(LED2_PIN     , HIGH); // off
  digitalWrite(SOLENOID4_PIN, LOW ); // off
  digitalWrite(SOLENOID3_PIN, LOW ); // off
  digitalWrite(SOLENOID2_PIN, LOW ); // off
  digitalWrite(SOLENOID1_PIN, LOW ); // off

  // finish configuring the LED hardware
  matrix.begin(ADA_HT1632_COMMON_16NMOS);
  matrix.fillScreen();
  matrix.clearScreen();
  matrix.writeScreen();
  strand.begin();
  strand.show();

  // initialize the Serial port for the user debugging console
  Serial.begin( BAUD_RATE );

  // send a message as a diagnostic
  send_debug_message("wakeup");
}

/****************************************************************/
/// Standard Arduino polling function to handle all I/O and periodic processing.
/// This function is called repeatedly as fast as possible from within the
/// built-in library to poll program events.  This loop should never be allowed
/// to stall or block so that all tasks can be constantly serviced.
void loop()
{
  // The timestamp in microseconds for the last polling cycle, used to compute
  // the exact interval between output updates.
  static unsigned long last_update_clock = 0;

  // Read the microsecond clock.
  unsigned long now = micros();

  // Compute the time elapsed since the last poll.  This will correctly handle wrapround of
  // the 32-bit long time value given the properties of twos-complement arithmetic.
  unsigned long interval = now - last_update_clock;
  last_update_clock = now;

  // Begin the polling cycle.
  poll_sensor_inputs(interval);
  poll_game_logic(interval);
  poll_actuator_outputs(interval);
  poll_console_input(interval);
}

/****************************************************************/
/****************************************************************/
